/*
Copyright (C) 2009

This file is part of the NN applet
written by Max Bügler
http://www.maxbuegler.eu/

NN applet is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

NN applet is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
package NN;

import java.util.Random;


/**
 * Author: Max Buegler - max@sarstech.com
 * Date: Dec 4, 2009
 * Time: 10:17:24 AM
 */
public class NeuralNetwork {
    private Matrix[] weights,bias,dw,db;
    private int[] structure;
    private static Random rnd=new Random();

    public NeuralNetwork(int[] structure) {
        this.structure=structure;
        weights=new Matrix[structure.length-1];
        bias=new Matrix[structure.length-1];
        dw=new Matrix[structure.length-1];
        db=new Matrix[structure.length-1];
        for (int x=0;x<structure.length-1;x++){
            int inputs=structure[x];
            int outputs=structure[x+1];
            weights[x]=createRandomMatrix(outputs,inputs);
            bias[x]=createRandomMatrix(outputs,1);
            dw[x]=createZeroMatrix(outputs,inputs);
            db[x]=createZeroMatrix(outputs,1);

        }
    }

    public Matrix feedForward(Matrix input, boolean hardlim){
        Matrix[] a=new Matrix[weights.length+1];
        a[0]=input;
        for (int x=1;x<weights.length;x++){
            a[x]=logsig(weights[x-1].times(a[x-1]).plus(bias[x-1]));
        }
        a[weights.length]=weights[weights.length-1].times(a[weights.length-1]).plus(bias[weights.length-1]);
        if (hardlim)return hardlim(a[weights.length]);
        return a[weights.length];
    }

    public Matrix backPropagate(Matrix input, Matrix target, double alpha, double gamma, boolean hardlim){
        Matrix[] a=new Matrix[weights.length+1];
        a[0]=input;
        for (int x=1;x<weights.length;x++){
            a[x]=logsig(weights[x-1].times(a[x-1]).plus(bias[x-1]));
        }
        a[weights.length]=weights[weights.length-1].times(a[weights.length-1]).plus(bias[weights.length-1]);
        Matrix err;
        if (hardlim)err=target.minus(hardlim(a[weights.length]));
        else err=target.minus(a[weights.length]);
        //BACK Propagation
        Matrix[] s=new Matrix[weights.length+1];
        s[weights.length-1]=eye(structure[structure.length-1],-2).times(err);
        for (int x=weights.length-2;x>=0;x--){
            s[x]=logsigderivative(a[x+1])
                    .times(weights[x+1].transpose())
                    .times(s[x+1]);
        }
        for (int x=0;x<weights.length;x++){
            Matrix deltaW=dw[x].times(gamma).plus(s[x].times(alpha).times(a[x].transpose()).times(1.0-gamma));
            weights[x]=weights[x].minus(deltaW);
            Matrix deltaB=db[x].times(gamma).plus(s[x].times(alpha).times(1.0-gamma));
            bias[x]=bias[x].minus(deltaB);
            dw[x]=deltaW;
            db[x]=deltaB;
        }
        //err.print();
        return err;
    }

    public void learnData(double[][] input, double[][] target, double alpha,double gamma,boolean hardlim){
        Matrix err=new Matrix(target[0].length,1);
        for (int x=0;x<input.length;x++){
            double[][] p=new double[input[x].length][1];
            for (int y=0;y<input[x].length;y++){
                p[y][0]=input[x][y];
            }
            double[][] t=new double[target[x].length][1];
            for (int y=0;y<target[x].length;y++){
                t[y][0]=target[x][y];
            }
            err=err.plus(backPropagate(new Matrix(p),new Matrix(t),alpha,gamma,hardlim));
        }
        err.times(1.0/input.length).print();
    }

    public Matrix[] getWeights(){
        return weights;
    }
    
    public Matrix[] getBias(){
        return bias;
    }


    private static Matrix logsig(Matrix in){
        double[][] out=new double[in.getH()][in.getW()];
        for (int x=0;x<in.getW();x++)
            for (int y=0;y<in.getH();y++)
                out[y][x]=1.0/(1+Math.exp(-in.getValue(y,x)));
        return new Matrix(out);
    }

    private static Matrix logsigderivative(Matrix a){
        double[][] out=new double[a.getH()][a.getH()];
        for (int x=0;x<out.length;x++){
            for (int y=0;y<out.length;y++){
                if (x==y)
                    out[x][y]=(1-a.getValue(x,0))*a.getValue(x,0);
                else
                    out[x][y]=0;
            }
        }
        return new Matrix(out);
    }

    private static Matrix createRandomMatrix(int r,int c){
        double[][] data=new double[r][c];
        for (int x=0;x<c;x++)
            for (int y=0;y<r;y++)
                data[y][x]=(2*rnd.nextDouble())-1;
        return new Matrix(data);
    }

    private static Matrix createZeroMatrix(int r,int c){
        double[][] data=new double[r][c];
        for (int x=0;x<c;x++)
            for (int y=0;y<r;y++)
                data[y][x]=0;
        return new Matrix(data);
    }

    private static Matrix eye(int s, double scale){
        double[][] out=new double[s][s];
        for (int x=0;x<s;x++){
            for (int y=0;y<s;y++){
                if (x==y)
                    out[y][x]=scale;
                else
                    out[y][x]=0;
            }
        }
        return new Matrix(out);
    }

    private static Matrix hardlim(Matrix in){
        double[][] out=new double[in.getH()][in.getW()];
        for (int x=0;x<in.getW();x++){
            for (int y=0;y<in.getH();y++){
                if (in.getValue(y,x)>=0.5)out[y][x]=1;
                else out[y][x]=0;
            }
        }
        return new Matrix(out);
    }

    public static void main(String[] args){
        NeuralNetwork nn=new NeuralNetwork(new int[]{2,2,2});
        double[][] p=new double[][]{
                {1,1},
                {1,0},
                {0,1},
                {0,0},
        };                                       //XOR Problem
        double[][] t=new double[][]{
                {0,1},
                {1,0},
                {1,0},
                {0,1},
        };
        for (int x=0;x<10000;x++){
            System.out.println("EPOCH "+x);
            nn.learnData(p,t,0.01,0,true);
        }
        System.out.println("OUTPUT for 1,1");
        nn.feedForward(new Matrix(new double[][]{{1},{1}}),true).print();
        System.out.println("OUTPUT for 1,0");
        nn.feedForward(new Matrix(new double[][]{{1},{0}}),true).print();
        System.out.println("OUTPUT for 0,1");
        nn.feedForward(new Matrix(new double[][]{{0},{1}}),true).print();
        System.out.println("OUTPUT for 0,0");
        nn.feedForward(new Matrix(new double[][]{{0},{0}}),true).print();

    }

    public void print(){
        for (int x=0;x<weights.length;x++){
            System.out.println("W"+(x+1)+": ");
            weights[x].print();
            System.out.println("B"+(x+1)+": ");
            bias[x].print();
        }
    }
    public String toString(){
        String out="[\t";
        for (int x=0;x<structure.length;x++){
            out+=structure[x]+"\t";
        }
        out+="]\n";
        for (int x=0;x<weights.length;x++){
            out+="W"+(x+1)+":\n";
            out+=weights[x].toString();
            out+="B"+(x+1)+":\n";
            out+=bias[x].toString();
        }
        return out;
    }

}
