/*
Copyright (C) 2009

This file is part of the NN applet
written by Max Bügler
http://www.maxbuegler.eu/

NN applet is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

NN applet is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
package Space;

public class Matrix {
    private final int h;             // number of rows
    private final int w;      // number of columns
    private final double[][] data;  

    /**
     * Create new Matrix
     * @param M rows
     * @param N columns
     */
    public Matrix(int M, int N) {
        this.h = M;
        this.w = N;
        data = new double[M][N];
    }

    /**
     * Create Matrix from 2D array
     * @param data 2D array
     */
    public Matrix(double[][] data) {
        h = data.length;
        w = data[0].length;
        this.data = new double[h][w];
        for (int i = 0; i < h; i++)
            for (int j = 0; j < w; j++)
                    this.data[i][j] = data[i][j];
    }
    /**
     * gets Value of position
     * @param x x-position
     * @param y y-position
     * @return value
     */
    public double getValue(int x, int y){
        return data[x][y];
    }

    
    private Matrix(Matrix A) { this(A.data); }

    private void swap(int i, int j) {
        double[] temp = data[i];
        data[i] = data[j];
        data[j] = temp;
    }

    /**
     * create and return the transpose of the invoking matrix
     * @return Matrix
     */
    public Matrix transpose() {
        Matrix A = new Matrix(w, h);
        for (int i = 0; i < h; i++)
            for (int j = 0; j < w; j++)
                A.data[j][i] = this.data[i][j];
        return A;
    }

    /**
     * Adds Matrix
     * @param B Matrix to add
     * @return Matrix
     */
    public Matrix plus(Matrix B) {
        Matrix A = this;
        if (B.h != A.h || B.w != A.w) throw new RuntimeException("Illegal matrix dimensions.");
        Matrix C = new Matrix(h, w);
        for (int i = 0; i < h; i++)
            for (int j = 0; j < w; j++)
                C.data[i][j] = A.data[i][j] + B.data[i][j];
        return C;
    }


    /**
     * Subtracts Matrix
     * @param B Matrix to subtract
     * @return Matrix
     */
    public Matrix minus(Matrix B) {
        Matrix A = this;
        if (B.h != A.h || B.w != A.w) throw new RuntimeException("Illegal matrix dimensions.");
        Matrix C = new Matrix(h, w);
        for (int i = 0; i < h; i++)
            for (int j = 0; j < w; j++)
                C.data[i][j] = A.data[i][j] - B.data[i][j];
        return C;
    }

    /**
     * Equals
     * @param B Matrix
     * @return Matrix
     */
    public boolean equals(Matrix B) {
        Matrix A = this;
        if (B.h != A.h || B.w != A.w) throw new RuntimeException("Illegal matrix dimensions.");
        for (int i = 0; i < h; i++)
            for (int j = 0; j < w; j++)
                if (A.data[i][j] != B.data[i][j]) return false;
        return true;
    }

    /**
     * mulitplies Matrix
     * @param B Matrix to multiply
     * @return Matrix
     */
    public Matrix times(Matrix B) {
        Matrix A = this;
        if (A.w != B.h) throw new RuntimeException("Illegal matrix dimensions.");
        Matrix C = new Matrix(A.h, B.w);
        for (int i = 0; i < C.h; i++)
            for (int j = 0; j < C.w; j++)
                for (int k = 0; k < A.w; k++)
                    C.data[i][j] += (A.data[i][k] * B.data[k][j]);
        return C;
    }


    /**
     * solves Matrix
     * return x = A^-1 b, assuming A is square and has full rank
     * @param rhs Matrix
     * @return Matrix
     */
    public Matrix solve(Matrix rhs) {
        if (h != w || rhs.h != w || rhs.w != 1)
            throw new RuntimeException("Illegal matrix dimensions.");

        // create copies of the data
        Matrix A = new Matrix(this);
        Matrix b = new Matrix(rhs);

        // Gaussian elimination with partial pivoting
        for (int i = 0; i < w; i++) {

            // find pivot row and swap
            int max = i;
            for (int j = i + 1; j < w; j++)
                if (Math.abs(A.data[j][i]) > Math.abs(A.data[max][i]))
                    max = j;
            A.swap(i, max);
            b.swap(i, max);

            // singular
            if (A.data[i][i] == 0.0) throw new RuntimeException("Matrix is singular.");

            // pivot within b
            for (int j = i + 1; j < w; j++)
                b.data[j][0] -= b.data[i][0] * A.data[j][i] / A.data[i][i];

            // pivot within A
            for (int j = i + 1; j < w; j++) {
                double m = A.data[j][i] / A.data[i][i];
                for (int k = i+1; k < w; k++) {
                    A.data[j][k] -= A.data[i][k] * m;
                }
                A.data[j][i] = 0.0;
            }
        }

        // back substitution
        Matrix x = new Matrix(w, 1);
        for (int j = w - 1; j >= 0; j--) {
            double t = 0.0;
            for (int k = j + 1; k < w; k++)
                t += A.data[j][k] * x.data[k][0];
            x.data[j][0] = (b.data[j][0] - t) / A.data[j][j];
        }
        return x;
   
    }
}
